from sql import util
from sql import store
from sql.exceptions import UsageError
from sql.cmd.cmd_utils import CmdParser
from sql.display import Table, Message
from sql.util import expand_args, is_rendering_required, render_string_using_namespace


def _modify_display_msg(key, remaining_keys, dependent_keys=None):
    """

    Parameters
    ----------
    key : str,
            deleted stored snippet
    remaining_keys: list
            snippets remaining after key is deleted
    dependent_keys: list
            snippets dependent on key

    Returns
    -------
    msg: str
        Formatted message
    """
    msg = f"{key} has been deleted.\n"
    if dependent_keys:
        msg = f"{msg}{', '.join(dependent_keys)} depend on {key}\n"
    if remaining_keys:
        msg = f"{msg}Stored snippets: {', '.join(remaining_keys)}"
    else:
        msg = f"{msg}There are no stored snippets"
    return msg


def snippets(others, user_ns):
    """
    Implementation of `%sqlcmd snippets`
    This function handles all the arguments related to %sqlcmd snippets, namely
    listing stored snippets, and delete/ force delete/ force delete a snippet and
    all its dependent snippets. It also uses the kernel namespace for expanding
    arguments declared as variables.


    Parameters
    ----------
    others : str,
        A string containing the command line arguments.

    user_ns : dict,
        User namespace of IPython kernel
    """
    parser = CmdParser()
    parser.add_argument(
        "-d", "--delete", type=str, help="Delete stored snippet", required=False
    )
    parser.add_argument(
        "-D",
        "--delete-force",
        type=str,
        help="Force delete stored snippet",
        required=False,
    )
    parser.add_argument(
        "-A",
        "--delete-force-all",
        type=str,
        help="Force delete all stored snippets",
        required=False,
    )
    all_snippets = store.get_all_keys()
    if len(others) == 1:
        others[0] = render_string_using_namespace(others[0], user_ns)
        if others[0] in all_snippets:
            return str(store.store[others[0]])

        base_err_msg = f"'{others[0]}' is not a snippet. "
        if len(all_snippets) == 0:
            err_msg = "%sThere is no available snippet."
        else:
            err_msg = "%sAvailable snippets are " f"{util.pretty_print(all_snippets)}."
        err_msg = err_msg % (base_err_msg)

        raise UsageError(err_msg)

    args = parser.parse_args(others)
    if is_rendering_required(" ".join(others)):
        expand_args(args, user_ns)

    SNIPPET_ARGS = [args.delete, args.delete_force, args.delete_force_all]
    if SNIPPET_ARGS.count(None) == len(SNIPPET_ARGS):
        if len(all_snippets) == 0:
            return Message("No snippets stored")
        else:
            return Table(["Stored snippets"], [[snippet] for snippet in all_snippets])

    if args.delete:
        deps = store.get_key_dependents(args.delete)
        if deps:
            deps = ", ".join(deps)
            raise UsageError(
                f"The following tables are dependent on {args.delete}: {deps}.\n"
                f"Pass --delete-force to only delete {args.delete}.\n"
                f"Pass --delete-force-all to delete {deps} and {args.delete}"
            )
        else:
            key = args.delete
            remaining_keys = store.del_saved_key(key)
            return _modify_display_msg(key, remaining_keys)

    elif args.delete_force:
        key = args.delete_force
        deps = store.get_key_dependents(key)
        remaining_keys = store.del_saved_key(key)
        for dep in deps:
            store.store[dep].remove_snippet_dependency(key)
        return _modify_display_msg(key, remaining_keys, deps)

    elif args.delete_force_all:
        deps = store.get_key_dependents(args.delete_force_all)
        deps.append(args.delete_force_all)
        for key in deps:
            remaining_keys = store.del_saved_key(key)
        return _modify_display_msg(", ".join(deps), remaining_keys)



from sql import exceptions
import sql.connection
from sqlglot import select, condition
from prettytable import PrettyTable
from sql.cmd.cmd_utils import CmdParser
from sql.util import expand_args, is_rendering_required


def return_test_results(args, conn, query):
    columns = []

    try:
        column_data = conn.execute(query).cursor.description
        res = conn.execute(query).fetchall()
        for column in column_data:
            columns.append(column[0])
        res = [columns, *res]
        return res
    except Exception as e:
        if "column" in str(e):
            raise exceptions.UsageError(
                f"Referenced column '{args.column}' not found!"
            ) from e


def run_each_individually(args, conn):
    if args.schema:
        table_ = f"{args.schema}.{args.table}"
    else:
        table_ = args.table
    base_query = select("*").from_(table_)

    storage = {}

    if args.greater:
        where = condition(args.column + "<=" + args.greater)
        current_query = base_query.where(where).sql()

        res = return_test_results(args, conn, query=current_query)

        if res is not None:
            storage["greater"] = res
    if args.greater_or_equal:
        where = condition(args.column + "<" + args.greater_or_equal)

        current_query = base_query.where(where).sql()

        res = return_test_results(args, conn, query=current_query)

        if res is not None:
            storage["greater_or_equal"] = res

    if args.less_than_or_equal:
        where = condition(args.column + ">" + args.less_than_or_equal)
        current_query = base_query.where(where).sql()

        res = return_test_results(args, conn, query=current_query)

        if res is not None:
            storage["less_than_or_equal"] = res
    if args.less_than:
        where = condition(args.column + ">=" + args.less_than)
        current_query = base_query.where(where).sql()

        res = return_test_results(args, conn, query=current_query)

        if res is not None:
            storage["less_than"] = res
    if args.no_nulls:
        where = condition("{} is NULL".format(args.column))
        current_query = base_query.where(where).sql()

        res = return_test_results(args, conn, query=current_query)

        if res is not None:
            storage["null"] = res

    return storage


def test(others, user_ns):
    """
    Implementation of `%sqlcmd test`

    This function takes in a string containing command line arguments,
    parses them to extract the table name, column name, and conditions
    to return if those conditions are satisfied in that table
    It also uses the kernel namespace for expanding arguments declared as
    variables.

    Parameters
    ----------
    others : str,
            A string containing the command line arguments.

    user_ns : dict,
        User namespace of IPython kernel

    Returns
    -------
    result: bool
        Result of the test

    table: PrettyTable
        table with rows because of which the test fails


    """
    parser = CmdParser()

    parser.add_argument("-t", "--table", type=str, help="Table name", required=True)
    parser.add_argument("-s", "--schema", type=str, help="Schema name", required=False)
    parser.add_argument("-c", "--column", type=str, help="Column name", required=False)
    parser.add_argument(
        "-g",
        "--greater",
        type=str,
        help="Greater than a certain number.",
        required=False,
    )
    parser.add_argument(
        "-goe",
        "--greater-or-equal",
        type=str,
        help="Greater or equal than a certain number.",
        required=False,
    )
    parser.add_argument(
        "-l",
        "--less-than",
        type=str,
        help="Less than a certain number.",
        required=False,
    )
    parser.add_argument(
        "-loe",
        "--less-than-or-equal",
        type=str,
        help="Less than or equal to a certain number.",
        required=False,
    )
    parser.add_argument(
        "-nn",
        "--no-nulls",
        help="Returns rows in specified column that are not null.",
        action="store_true",
    )

    args = parser.parse_args(others)
    if is_rendering_required(" ".join(others)):
        expand_args(args, user_ns)

    COMPARATOR_ARGS = [
        args.greater,
        args.greater_or_equal,
        args.less_than,
        args.less_than_or_equal,
    ]

    if args.table and not any(COMPARATOR_ARGS):
        raise exceptions.UsageError("Please use a valid comparator.")

    if args.table and any(COMPARATOR_ARGS) and not args.column:
        raise exceptions.UsageError("Please pass a column to test.")

    if args.greater and args.greater_or_equal:
        return exceptions.UsageError(
            "You cannot use both greater and greater "
            "than or equal to arguments at the same time."
        )
    elif args.less_than and args.less_than_or_equal:
        return exceptions.UsageError(
            "You cannot use both less and less than "
            "or equal to arguments at the same time."
        )

    conn = sql.connection.ConnectionManager.current
    result_dict = run_each_individually(args, conn)

    if any(len(rows) > 1 for rows in list(result_dict.values())):
        for comparator, rows in result_dict.items():
            if len(rows) > 1:
                print(f"\n{comparator}:\n")
                _pretty = PrettyTable()
                _pretty.field_names = rows[0]
                for row in rows[1:]:
                    _pretty.add_row(row)
                print(_pretty)
        raise exceptions.UsageError(
            "The above values do not match your test requirements."
        )
    else:
        return True



import os
from io import open
import re
import ast

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, "README.md"), encoding="utf-8").read()

_version_re = re.compile(r"__version__\s+=\s+(.*)")

with open("src/sql/__init__.py", "rb") as f:
    VERSION = str(
        ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    )

install_requires = [
    "prettytable>=3.12.0",
    # IPython dropped support for Python 3.8
    "ipython<=8.12.0; python_version <= '3.8'",
    "sqlalchemy",
    "sqlparse",
    "ipython-genutils>=0.1.0",
    "jinja2",
    "sqlglot>=11.3.7",
    'importlib-metadata;python_version<"3.8"',
    # we removed the share notebook button in this version
    "jupysql-plugin>=0.4.2",
    "ploomber-core>=0.2.7",
]

DEV = [
    "flake8",
    "pytest",
    # 24/01/24 Pandas 2.2.0 breaking CI: https://github.com/ploomber/jupysql/issues/983
    "pandas<2.2.0",  # previously pinned to 2.0.3
    "polars==0.17.2",  # 04/18/23 this breaks our CI
    "pyarrow",
    "invoke",
    "pkgmt",
    "twine",
    # tests
    "duckdb<1.1.0",
    "duckdb-engine",
    "pyodbc",
    # sql.plot module tests
    "matplotlib==3.7.2",
    "black",
    # for %%sql --interact
    "ipywidgets",
    # for running tests for %sqlcmd explore --table
    "js2py",
    # for monitoring access to files
    "psutil",
    # for running tests for %sqlcmd connect
    "jupyter-server",
]

# dependencies for running integration tests
INTEGRATION = [
    "dockerctx",
    "pyarrow",
    "psycopg2-binary",
    "pymysql",
    "pgspecial==2.0.1",
    "pyodbc",
    "snowflake-sqlalchemy",
    "oracledb",
    "sqlalchemy-pytds",
    "python-tds",
    # redshift
    "redshift-connector",
    "sqlalchemy-redshift",
    "clickhouse-sqlalchemy",
    # following two dependencies required for spark
    "pyspark",
    "grpcio-status",
]

setup(
    name="jupysql",
    version=VERSION,
    description="Better SQL in Jupyter",
    long_description=README,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Database",
        "Topic :: Database :: Front-Ends",
        "Programming Language :: Python :: 3",
    ],
    keywords="database ipython postgresql mysql duckdb",
    author="Ploomber",
    author_email="contact@ploomber.io",
    url="https://github.com/ploomber/jupysql",
    project_urls={
        "Source": "https://github.com/ploomber/jupysql",
    },
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        "dev": DEV,
        "integration": DEV + INTEGRATION,
    },
)
